import traceback

from scopeton import DiTools
from scopeton.scopetonException import ScopetonException


def InjectClass(**kwargs_):
    def decorator(cls):
        def inject(self):
                for name in kwargs_:
                    try:
                        setattr(self, name, self.TTTcontextScope.getInstance(kwargs_[name]))
                    except Exception as e:
                        raise ScopetonException("Error while injecting {name} into {cls}, type is: {obj}, nested exception: {err}".format(name=name, cls = cls, obj = kwargs_[name], err=traceback.format_exc()), e)

        cls.TTTinjectMethod = inject
        return cls
    return decorator

def PostConstruct():
    """
    method marked by postConstruct will be called on bean initialization
    :return:
    """
    def decorator(fn):
        def initAnnotation(self):
            fn(self)
        DiTools.beanAnnotateMethod(fn, initAnnotation)
        return fn
    return decorator

def PreDestroy():
    def decorator(fn):
        def initAnnotation(self):
            self.TTTPreDestroy = fn
        DiTools.beanAnnotateMethod(fn, initAnnotation)
        return fn
    return decorator


import copy
from threading import RLock

import scopeton.DiTools as DiTools
from scopeton.ContextBean import ContextBean

lock = RLock()


def _StaticSync(fn):
    def wrapped(*args, **kwargs):
        lock.acquire()
        try:
            return fn(*args, **kwargs)
        finally:
            lock.release()

    return wrapped


class StaticContext(object):
    _beans = {}  # type: dict[str, object]

    @classmethod
    @_StaticSync
    def getBeansCopy(cls):
        return copy.copy(cls._beans)

    @classmethod
    @_StaticSync
    def registerBean(cls, clazz, lazy=True, name=None):
        bean = DiTools.createBean(clazz, lazy=lazy)
        cls._beans[name or bean.name] = bean

def Service(lazy=True, name=None):
    def decorator(cls):
        StaticContext.registerBean(cls, lazy=lazy, name=name)
        return cls

    return decorator



class ContextBean(object):
    name = ""
    object = None  # type: object
    postConstruct = None
    preDestroy = None
    lazy = False
    def __str__(self):
        return "ContextBean[{self.name}](lazy={self.lazy})".format(self=self)
    def __repr__(self):
        return self.__str__()


from threading import RLock

import scopeton.DiTools as DiTools
from scopeton.ContextBean import ContextBean

class ScopeContext(object):
    def __init__(self, beans, servicesAutoStart=True):
        # type: (dict[str, ContextBean], boolean) -> object
        self.__instances = {}  # type: dict[str, object]
        self.scopeLock = RLock()
        self.beans = beans  # type: (dict[str, ContextBean]) -> object
        if servicesAutoStart:
            self.startServices()

    def startServices(self):
        for serviceQualifier in self.beans:
            if not self.beans[serviceQualifier].lazy:
                self.getInstance(self.beans[serviceQualifier].object)

    def resolveClass(self, name):
        # type: (object) -> ContextBean
        pkg = DiTools.getFullyQualifiedName(name)
        if pkg not in self.beans:
            raise Exception("Error, no class with name {pkg} registered, available:{keys}".format(pkg=pkg, keys=self.beans.keys()))
        return self.beans[pkg]

    def __inject(self, instance):
        if hasattr(instance, "TTTinjectMethod"):
            instance.TTTinjectMethod()

    def __initInstance(self, instance):
        instance.TTTcontextScope = self
        annotatedMethods = DiTools.getBeanMethodsInitializers(instance.__class__)
        for method in annotatedMethods:
            for annotationInitFunction in annotatedMethods[method]:
                annotationInitFunction(instance)
        return instance

    def _makeInstance(self, cls):
        # type: (ContextBean) -> object
        return self.__initInstance(cls.object())

    def registerInstance(self, name, instance):
        self.__instances[DiTools.getFullyQualifiedName(name)] = instance
        return self.__initInstance(instance)

    def registerBean(self, name, cls):
        self.beans[DiTools.getFullyQualifiedName(name)] = DiTools.createBean(cls)

    def getInstance(self, obj):
        name = DiTools.getFullyQualifiedName(obj)
        if name in self.__instances:
            return self.__instances[name]
        cls = self.resolveClass(obj)
        self.scopeLock.acquire()
        try:
            if name not in self.__instances:
                self.__instances[name] = self._makeInstance(cls)
                self.__inject(self.__instances[name])
        finally:
            self.scopeLock.release()
        return self.__instances[name]

    def destroyInstance(self, instance):
        if hasattr(instance, "TTTPreDestroy"):
            instance.TTTPreDestroy()

    def destroy(self):
        self.scopeLock.acquire()
        try:
            for name in self.__instances:
                instance = self.__instances[name]
                self.destroyInstance(instance)
        finally:
            self.scopeLock.release()


import inspect
from threading import RLock
import sys

from scopeton.ContextBean import ContextBean

cacheLock = RLock()
def getSimpleNameFromString(obj):
    return obj.split(".")[-1]

def __getSimpleClassNameFromObjectInner(obj):
    """returns class name (without package for given object)"""
    # type: (object) -> str
    tp = type(obj)
    if tp.__name__ == 'classobj':
         return getSimpleNameFromString(obj.__name__)
    elif tp is inspect.isclass(obj) or tp is type:
        return getSimpleNameFromString(obj.__name__)
    elif tp is str:
        return obj
    elif (hasattr(obj, "__class__") and hasattr(obj.__class__, "__name__")):
        return getSimpleNameFromString(obj.__class__.__name__)
    raise Exception("Cannot get className for object:" + str(type(obj)))


def getSimpleClassNameFromObject(obj):
    """returns class name (without package for given object)"""
    # type: (object) -> str
    cacheLock.acquire()
    try:
        if not hasattr(obj, "__simpleClassName"):
            obj.__simpleClassName = __getSimpleClassNameFromObjectInner(obj)
    finally:
        cacheLock.release()
    return obj.__simpleClassName



def getObjectPackage(obj):
    # type: (object) -> str
    """returns object package"""
    if type(obj) is str:
        return obj
    if not hasattr(obj, "__module__"):
        if type(obj).__name__ == 'module':
            possibleError = "Type to inject is module, mb you imported/injected module instead of class: {obj}".format(obj=obj)
        else:
            possibleError = "unknown"

        raise Exception("Error, object has no attribute __module__:{obj}, type:{type}, possible error:{possibleError}".format(obj=obj, type=type(obj), possibleError = possibleError))
    return obj.__module__


def __getFullyQualifiedNameInner(obj):
    # type: (object) -> str
    """returns class name with package"""
    result = []
    package = getObjectPackage(obj)
    if package:
        result.append(package)
    result.append(getSimpleClassNameFromObject(obj))
    return ".".join(result)

def getFullyQualifiedName(obj):
    # type: (object) -> str
    """returns class name with package"""
    if type(obj) is str:
        return obj
    cacheLock.acquire()
    try:
        if not hasattr(obj, "__fullyQualifiedClassName"):
            obj.__fullyQualifiedClassName = __getFullyQualifiedNameInner(obj)
    finally:
        cacheLock.release()
    return obj.__fullyQualifiedClassName

def getScope(scopeHolder):
    if hasattr(scopeHolder, "TTTcontextScope"):
        return scopeHolder.TTTcontextScope
    raise Exception("Error, object is not scopeHolder:" + (str(scopeHolder)))

def getClassMethods(cls):
    # type: (object) -> list[object]
    if sys.version_info[0] > 2:
        methods = inspect.getmembers(cls, predicate=inspect.ismethod) + inspect.getmembers(cls, predicate=inspect.isfunction)
        methods = map(lambda meth: meth[1], methods)
        return list(methods)
    else:
        return map(lambda meth: meth[1], inspect.getmembers(cls, predicate=inspect.ismethod))

def beanAnnotateMethod(method, annotation):
    cacheLock.acquire()
    try:
        if not hasattr(method, "TTTannotations"):
            method.TTTannotations = []
        if annotation in method.TTTannotations:
            raise Exception("Error, method: {methodName} already has annotation: {annotationName}".format(methodName=method.__name__,annotationName=annotation.__name__))
        method.TTTannotations.append(annotation)
    finally:
        cacheLock.release()

def getBeanMethodsInitializers(clazz):
    # type: (object) -> dict[object, object]
    cacheLock.acquire()
    try:
        if hasattr(clazz, "__annotatedMethods"):
            return clazz.__annotatedMethods
        methods = {}
        for method in getClassMethods(clazz):
            if hasattr(method, "TTTannotations"):
                methods[method] = method.TTTannotations
        clazz.__annotatedMethods = methods
        return methods
    finally:
        cacheLock.release()

def createBean(clazz, lazy=True):
    bean = ContextBean()
    bean.lazy = lazy
    bean.name = getFullyQualifiedName(clazz)
    bean.object = clazz
    return bean


import traceback

class ScopetonException(Exception):
    errorText = ""
    parentError = None  # type: Exception

    def __init__(self, errorText, parentError):
        self.errorText = errorText
        self.parentError = parentError
        self.parentStack = traceback.format_exc()

    def __str__(self):
        return "ScopetonException:{e.errorText}\nNested exception:{e.parentError}\n{e.parentStack}".format(e=self)
